/*
 * Beispielprogramm eines TWI (I2C) Masters. Getestet mit ATMega8
 *
 * Die gesamte Behandlung der TWI-Zustnde erfolgt in der Interrupt-Routine. Das Hauptprogramm startet nur
 * den Transfer.
 *
 */

#include "util/delay.h"
#include "avr/interrupt.h"
#include "util/twi.h"

volatile uint8_t busy;			// Busy-Flag, gesetzt, so lange ein Transfer luft.
volatile uint8_t slaRW;			// Slave-Adresse und R/W-Kennzeichen. Gesetzt im Hauptprogramm
static uint8_t data[64];		// Datenpuffer
volatile uint8_t ind;			// Index in den Datenpuffer
static uint8_t max;				// Empfangslnge. Wird im ersten Byte vom Slave bei der Betriebsart Master-Receiver
								// gesetzt.

/*
 * Die einzelnen Zustnde der TWI-State-machine findet man in der Atmel Dokumentation.
 * Der Kommentar gibt die Ursache des Interrupts an.
 */
ISR(TWI_vect) {
	switch (TWSR & TW_STATUS_MASK) {
	// Start oder wiederholten Start gesendet.
	case TW_REP_START:
	case TW_START:
		ind = 0;
		TWDR = slaRW;
		TWCR = (1 << TWINT) | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		break;
	// TW_MT.... Master Transmitter
	// Slave hat Adresse oder Datenbyte mit ACK besttigt.
	case TW_MT_SLA_ACK:
	case TW_MT_DATA_ACK:
		TWDR = data[ind++];
		if (ind <= 16) {
			// es sind noch bytes zu senden
			TWCR = (1 << TWINT) | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		} else {
			// alle Bytes gesendet, Stop erzeugen.
			TWCR = (1 << TWINT) | (0 << TWEA) | (0 << TWSTA) | (1 << TWSTO) | (1 << TWEN) | (1 << TWIE);
			busy = 0;
		}
		break;
	// Arbitration losses bei mehreren Mastern werden nicht behandelt.
	case TW_MR_ARB_LOST:
		break;
	// TW_MR.... Master Receiver
	// Slave hat Adresse erkannt und besttigt. Index auf 0, max auf 0 fr Erkennung 1. Byte.
	case TW_MR_SLA_ACK:
		TWCR = (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		ind = 0;
		max = 0;
		break;
	// Daten sind mit ACK eingetroffen
	case TW_MR_DATA_ACK:
		if (max == 0) {
			// Erstes Byte ist die Lnge der Nachricht.
			max = TWDR;
		} else {
			// Datenbyte speichern
			data[ind++] = TWDR;
		}
		if (ind < (max - 1)) {
			// noch mindestens zwei Byte zu bertragen: ACK
			TWCR = (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		} else {
			// Vorletztes Byte: NACK
			TWCR = (1 << TWINT) | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		}
		break;
	// Datenbyte mit NACK erhalten: letztes Byte ablegen, STOP erzeugen
	case TW_MR_DATA_NACK:
		data[ind++] = TWDR;
	// Keine Adressquittung: STOP erzeugen
	case TW_MR_SLA_NACK:
		TWCR = (1 << TWINT) | (0 << TWEA) | (0 << TWSTA) | (1 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		busy = 0;
		break;
	}
}

int main(int argc, char** argv) {
	uint8_t senden;
	uint8_t i;

	DDRB = 0x00;
	PORTB = 0xff;
	DDRC = ((0 << DDC0) | (0 << DDC1) | (0 << DDC2) | (0 << DDC3) | (1 << DDC4) | (1 << DDC5));
	PORTC = ((1 << DDC0) | (1 << DDC1) | (1 << DDC2) | (1 << DDC3) | (0 << DDC4) | (0 << DDC5));
	DDRD = 0x00;
	PORTD = 0xff;

	// Busrate setzen. TWBR soll > 10 sein, Prescaler (TWSR) wird auf 1 gesetzt
	TWBR = 16;
	TWSR = 0;
	busy = 0;
	sei();

	for (i = 0; i < sizeof data; i++) {
		data[i] = 0x40 + i;
	}

	// abwechselnd 16 Byte senden und 8 Byte (s. Slave) empfangen.
	senden = 1;
	for (;;) {
		if (!busy) {
			busy = 1;
			// dieser Delay scheint beim Wechsel der Betriebsart notwendig zu sein.
			_delay_loop_1(20);
			if (senden) {
				senden = 0;
				// Adresse 1, Schreiben
				slaRW = (1 << 1) | TW_WRITE;
				// Startbedingung erzeugen
				TWCR = (1 << TWINT) | (0 << TWEA) | (1 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
			} else {
				senden = 1;
				// Adresse 1, Lesen
				slaRW = (1 << 1) | TW_READ;
				// Startbedingung erzeugen
				TWCR = (1 << TWINT) | (0 << TWEA) | (1 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
			}
		}
	}
}

