/*
 * Beispielprogramm eines TWI (I2C) Slave. Getestet mit ATMega8
 *
 * Die gesamte Behandlung der TWI-Zustnde erfolgt in der Interrupt-Routine. Das Hauptprogramm tut nichts.
 * Wenn die Daten nicht direkt in der Interruptroutine bearbeitet werden knnen, muss man am Ende den Interrupt ausschalten
 * und nach der Verarbeitung im Haptprogramm wieder einschalten. Aber Achtung: wenn der Prozessor nicht auf TWI-Adressanfragen reagiert,
 * gibt es einen SLA_NACK beim Sender.
 *
 */

#include "util/delay.h"
#include "avr/interrupt.h"
#include "util/twi.h"

static uint8_t data[64];		// Datenpuffer
volatile uint8_t ind;			// Index in den Datenpuffer
volatile uint8_t max;			// Anzahl der zu sendenden Bytes bei Slave-Transmitter-Betrieb

/*
 * Die einzelnen Zustnde der TWI-State-machine findet man in der Atmel Dokumentation.
 * Der Kommentar gibt die Ursache des Interrupts an.
 */
ISR(TWI_vect) {
	switch (TWSR & TW_STATUS_MASK) {
	// TW_SR.... Slave Receiver
	// Erkennung der eigene Adresse nach verschiedenen Startbedingungen
	case TW_SR_ARB_LOST_SLA_ACK:
	case TW_SR_ARB_LOST_GCALL_ACK:
	case TW_SR_GCALL_ACK:
	case TW_SR_SLA_ACK:
		ind = 0;
		// Adresse quittieren
		TWCR = (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		break;
	// Datenbyte erkannt und mit ACK quittiert
	case TW_SR_GCALL_DATA_ACK:
	case TW_SR_DATA_ACK:
		// Noch Platz im Puffer?
		if (ind < sizeof data) {
			data[ind++] = TWDR;
		}
		// Quittieren und nchstes Byte anfordern
		TWCR = (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		break;
	// NACK im Receiver-Mode fhrt zum Rcksetzen des TWI-Interfaces
	case TW_SR_GCALL_DATA_NACK:
	case TW_SR_DATA_NACK:
		TWCR = (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		ind = 0;
		break;
	// STOP erkannt: TWI-Interface in Grundstellung bringen.
	case TW_SR_STOP:
		TWCR = (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		ind = 0;
		break;
	// TW_ST.... Slave Transmitter
	// Adresse erkannt, erstes Byte (Lnge) senden
	case TW_ST_ARB_LOST_SLA_ACK:
	case TW_ST_SLA_ACK:
		ind = 0;
		max = 8;		// Lnge des Sendetelegramms
		TWDR = max;
		TWCR = (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		break;
	// Byte wurde quittiert, nchstes Byte ist angefordert
	case TW_ST_DATA_ACK:
		if (ind < max) {
			// Es sind noch Daten da, also ACK
			TWDR = data[32 + ind++];
			TWCR = (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		} else {
			// Zu viele Daten angefordert, also NACK
			TWCR = (1 << TWINT) | (0 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		}
		break;
	// NACK bzw. ACK auf letztes Datenbyte erhalten: TWI-Interface in Grundstellung
	case TW_ST_DATA_NACK:
	case TW_ST_LAST_DATA:
		// hier die Daten verarbeiten. Z.B. umkopieren o.. Wenn es lngere Aktionen sind, in das Hauptprogramm auslagern
		TWCR = (1 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
		ind = 0;
		break;
	}
}

int main(int argc, char** argv) {
	uint8_t i;

	DDRB = 0x00;
	PORTB = 0xff;
	DDRC = ((0 << DDC0) | (0 << DDC1) | (0 << DDC2) | (0 << DDC3) | (1 << DDC4) | (1 << DDC5));
	PORTC = ((1 << DDC0) | (1 << DDC1) | (1 << DDC2) | (1 << DDC3) | (0 << DDC4) | (0 << DDC5));
	DDRD = 0x00;
	PORTD = 0xff;

	for (i = 0; i < sizeof data; i++) {
		data[i] = i;
	}

	// Adressregister (TWAR) auf Adresse setzen. Achtung: 1x links schieben, da Bit 0 fr Global call verwendet wird
	TWAR = (1 << 1);
	// Interrupt erlauben und ACK auf Adresse aktivieren
	TWCR = (0 << TWINT) | (1 << TWEA) | (0 << TWSTA) | (0 << TWSTO) | (1 << TWEN) | (1 << TWIE);
	sei();

	// alles luft jetzt im Interrupt
	for (;;) {
	}
}
